@@ -1,6 +1,6 @@
 /*
 ** Splint - annotation-assisted static program checker
-** Copyright (C) 1994-2002 University of Virginia,
+** Copyright (C) 1994-2003 University of Virginia,
 **         Massachusetts Institute of Technology
 **
 ** This program is free software; you can redistribute it and/or modify it
@@ -37,6 +37,8 @@
  * Herbert 06/12/2000:
  * - added OS/2 specific includes before osd_getPid()
  * - handle files like in WIN32 for OS/2 in osd_fileExists()
+ * Herbert 02/17/2002:
+ * - added OS/2 support to absolute file names
  */
 
 /*@-allmacros*/
@@ -45,14 +47,12 @@
 # include <sys/stat.h>
 /* Fix suggested by Lars Rasmussen */
 # include <errno.h>
-# ifdef WIN32
-# include <direct.h>
-# define getcwd _getcwd
-# endif
 
 /* POSIX platforms should defined getpid in unistd.h */
 # if defined (WIN32) || (defined(OS2) && defined(__IBMC__))
 # include <process.h>
+# include <direct.h>
+# define getcwd _getcwd
 # else
 # include <unistd.h>
 # endif
@@ -62,7 +62,6 @@
 # include "splintMacros.nf"
 # include "basic.h"
 # include "osd.h"
-# include "portab.h"
 
 /* from stat.h */
 /*@ignore@*/
@@ -184,7 +183,7 @@ osd_getPath (cstring path, cstring file, cstring *returnPath)
       *fullPath == '\0' || 
       (*file == CONNECTCHAR || (file[0] != '\0' && file[1] == ':'))
 # else
-    (*file == CONNECTCHAR)
+      (*file == CONNECTCHAR)
 # endif
       )
     {
@@ -204,8 +203,8 @@ osd_getPath (cstring path, cstring file, cstring *returnPath)
       /* Path specified. Loop through directories in path looking for the */
       /* first occurrence of the file.				    */
       
-      while (nextdir (&fullPath, &dirPtr, &dirLen) &&
-	     rVal == OSD_FILENOTFOUND)
+      while (nextdir (&fullPath, &dirPtr, &dirLen) 
+	     && rVal == OSD_FILENOTFOUND)
 	{
 	  if ((dirLen + strlen (file) + 2) <= MAXPATHLEN)
 	    {
@@ -305,7 +304,7 @@ osd_fileExists (cstring filespec)
   return (stat (cstring_toCharsSafe (filespec), &buf) == 0);
 # else
 # if defined (WIN32) || defined (OS2)
-  FILE *test = fileTable_openFile (context_fileTable (), filespec, "r");
+  FILE *test = fileTable_openReadFile (context_fileTable (), filespec);
   
   if (test != NULL) 
     {
@@ -453,8 +452,6 @@ nextdir (d_char *current_dir, d_char *dir, size_t *len)
 # endif
 }
 
-# ifndef NOLCL
-
 # ifdef WIN32
 extern /*@external@*/ int _flushall (void) /*@modifies fileSystem@*/ ;
 # endif
@@ -473,7 +470,6 @@ int osd_system (cstring cmd)
   res = system (cstring_toCharsSafe (cmd));
   return res;
 }
-# endif
 
 # ifndef unlink
 /* This should be defined by unistd.h */
@@ -482,7 +478,7 @@ extern /*@external@*/ int unlink (const char *) /*@modifies fileSystem@*/ ;
 /*@=redecl@*/
 # endif
 
-static s_tempError = FALSE;
+static bool s_tempError = FALSE;
 
 void osd_setTempError (void)
 {
@@ -567,7 +563,7 @@ cstring osd_fixDefine (cstring x)
 
 bool osd_fileIsReadable (cstring f)
 {
-  FILE *fl = fileTable_openFile (context_fileTable (), f, "r");
+  FILE *fl = fileTable_openReadFile (context_fileTable (), f);
 
   if (fl != NULL)
     {
@@ -776,7 +772,7 @@ osd_dirAbsolute (char *str)
   
   DPRINTF (("Absolute for: %s", str));
 
-# ifdef WIN32
+# if defined (WIN32) || defined (OS2) || defined (MSDOS)
   if (strlen (str) > 1 && str[1] == ':')
     {
       /*
@@ -861,7 +857,7 @@ static /*@only@*/ cstring osd_cwd = cstring_undefined;
 
 static void osd_setWorkingDirectory (void)
 {
-# ifdef UNIX
+# if defined (UNIX) || defined (OS2)
   char *buf = dmalloc (sizeof (*buf) * MAXPATHLEN);
   char *cwd = getcwd (buf, MAXPATHLEN);
 
@@ -871,6 +867,7 @@ static void osd_setWorkingDirectory (void)
     {
       lldiagmsg (message ("Cannot get working directory: %s\n",
 			  lldecodeerror (errno)));
+      osd_cwd = cstring_makeLiteral ("<missing directory>");
     }
   else
     {
@@ -888,9 +885,15 @@ void osd_initMod (void)
   osd_setWorkingDirectory ();
 }
 
+void osd_destroyMod (void)
+{
+  cstring_free (osd_cwd);
+  osd_cwd = cstring_undefined;
+}
+
 cstring osd_absolutePath (cstring cwd, cstring filename)
 {
-# ifdef UNIX
+# if defined (UNIX) || defined (OS2)
   /* Setup the current working directory as needed.  */
   cstring cwd2 = cstring_isDefined (cwd) ? cwd : osd_cwd;
   char *abs_buffer;
@@ -900,7 +903,7 @@ cstring osd_absolutePath (cstring cwd, cstring filename)
   llassert (cstring_isDefined (cwd2));
   llassert (cstring_isDefined (filename));
 
-  abs_buffer = (char *) dmalloc (size_fromInt (cstring_length (cwd2) + cstring_length (filename) + 2));
+  abs_buffer = (char *) dmalloc (cstring_length (cwd2) + cstring_length (filename) + 2);
   endp = abs_buffer;
   
   /*
@@ -911,7 +914,11 @@ cstring osd_absolutePath (cstring cwd, cstring filename)
   {
     const char *src_p;
 
-    if (filename[0] != '/')
+    if (!osd_isConnectChar (filename[0])
+# ifdef OS2
+	&& !(isalpha (filename[0]) && filename[1] == ':')
+# endif
+	)
       {
         src_p = cwd2;
 
@@ -920,7 +927,7 @@ cstring osd_absolutePath (cstring cwd, cstring filename)
 	    continue;
 	  }
 
-        *(endp-1) = '/';                        /* overwrite null */
+        *(endp-1) = CONNECTCHAR;                        /* overwrite null */
       }
 
     src_p = filename;
@@ -946,29 +953,29 @@ cstring osd_absolutePath (cstring cwd, cstring filename)
 	{
 	  break;
 	}
-      else if (inp[0] == '/' && outp[-1] == '/')
+      else if (osd_isConnectChar (inp[0]) && osd_isConnectChar (outp[-1]))
 	{
 	  inp++;
 	  continue;
 	}
-      else if (inp[0] == '.' && outp[-1] == '/')
+      else if (inp[0] == '.' && osd_isConnectChar (outp[-1]))
 	{
 	  if (inp[1] == '\0')
 	    {
 	      break;
 	    }
-	  else if (inp[1] == '/')
+	  else if (osd_isConnectChar (inp[1]))
 	    {
 	      inp += 2;
 	      continue;
 	    }
 	  else if ((inp[1] == '.') 
-		   && (inp[2] == '\0' || inp[2] == '/'))
+		   && (inp[2] == '\0' || osd_isConnectChar (inp[2])))
 	    {
-	      inp += (inp[2] == '/') ? 3 : 2;
+	      inp += (osd_isConnectChar (inp[2])) ? 3 : 2;
 	      outp -= 2;
 	
-	      while (outp >= abs_buffer && *outp != '/')
+	      while (outp >= abs_buffer && !osd_isConnectChar (*outp))
 		{
 		  outp--;
 		}
@@ -1002,7 +1009,7 @@ cstring osd_absolutePath (cstring cwd, cstring filename)
      the last character of the returned string is *not* a slash.  */
   
   *outp = '\0';
-  if (outp[-1] == '/')
+  if (osd_isConnectChar (outp[-1]))
     *--outp  = '\0';
   
   /*@noaccess cstring@*/
@@ -1029,19 +1036,28 @@ cstring osd_absolutePath (cstring cwd, cstring filename)
 
 cstring osd_outputPath (cstring filename)
 {
-# ifdef UNIX
+# if defined (UNIX) || defined (OS2)
   char *rel_buffer;
   char *rel_buf_p;
   cstring cwd_p = osd_cwd;
   char *path_p;
   int unmatched_slash_count = 0;
-  size_t filename_len = size_fromInt (cstring_length (filename));
+  size_t filename_len = cstring_length (filename);
   
+  llassertretval (filename_len > 0, filename);
+
   /*@access cstring@*/
   path_p = filename;
-  rel_buf_p = rel_buffer = (char *) dmalloc (filename_len);
+  rel_buffer = (char *) dmalloc (filename_len);
+  rel_buf_p = rel_buffer;
   *rel_buf_p = '\0';
 
+  if (cwd_p == NULL) 
+    {
+      /* Need to prevent recursive assertion failures */
+      return cstring_copy (filename);
+    }
+
   llassert (cwd_p != NULL);
   llassert (path_p != NULL);
 
@@ -1051,28 +1067,37 @@ cstring osd_outputPath (cstring filename)
       path_p++;
     }
   
-  if ((*cwd_p == '\0') && (*path_p == '\0' || *path_p == '/'))  /* whole pwd matched */
+  if ((*cwd_p == '\0') && (*path_p == '\0' || osd_isConnectChar (*path_p)))  /* whole pwd matched */
     {
       if (*path_p == '\0')             /* input *is* the current path! */
-	return cstring_makeLiteral (".");
+	{
+	  cstring_free (rel_buffer);
+	  return cstring_makeLiteral (".");
+	}
       else
 	{
-	  /*@i324 ! splint didn't report an errors for: return ++path_p; */
-	  return cstring_fromCharsNew (++path_p);
+	  cstring_free (rel_buffer);
+	  return cstring_fromCharsNew (path_p + 1);
 	}
     }
   else
     {
+
+      /* drl   2002-10/14 I had to put this code back*/
+      /* the case that needs it is when splint is given an absolute path name of a file outside of the current directory and the subdirectories below the current directory. e.g. cd /home/; splint /tmp/prog.c
+       */
+      
       /* evans 2002-02-05 This is horrible code, which I've removed.  I couldn't find any
       ** test cases that need it, so I hope I'm not breaking anything.
       */
+      /*#if 0*/
 
-# if 0      
       if (*path_p != '\0')
         {
           --cwd_p;
           --path_p;
-          while (*cwd_p != '/')         /* backup to last slash */
+
+          while (cwd_p >= osd_cwd && !osd_isConnectChar (*cwd_p)) /* backup to last slash */
             {
               --cwd_p;
               --path_p;
@@ -1085,7 +1110,7 @@ cstring osd_outputPath (cstring filename)
       /* Find out how many directory levels in cwd were *not* matched.  */
       while (*cwd_p != '\0')
 	{
-	  if (*cwd_p++ == '/')
+	  if (osd_isConnectChar (*cwd_p++))
 	    unmatched_slash_count++;
 	}
       
@@ -1093,16 +1118,28 @@ cstring osd_outputPath (cstring filename)
          Reject it if longer than the input.  */
       if (unmatched_slash_count * 3 + strlen (path_p) >= filename_len)
 	{
+	  cstring_free (rel_buffer);
+	  /* fprintf (stderr, "Returning filename: %s [%p]\n", filename); */
 	  return cstring_copy (filename);
 	}
-# endif
-      
+
+      /*drl 10-14-2002 end previously removed code */
+      /*#endif*/
       /* For each of them, put a `../' at the beginning of the short name.  */
       while (unmatched_slash_count-- > 0)
         {
           /* Give up if the result gets to be longer
              than the absolute path name.  */
-          if (rel_buffer + filename_len <= rel_buf_p + 3)
+	  char * temp_rel_buf_p;
+
+	  /*drl This comment is necessary because for some reason Splint
+	    does not realize that the pasts where rel_buf_p is released
+	    do not reach here*/
+	  /*@-usereleased@*/
+	  temp_rel_buf_p = rel_buf_p;
+	  /*@-usereleased@*/
+	  
+          if (rel_buffer + filename_len <= temp_rel_buf_p + 3)
 	    {
 	      sfree (rel_buffer);
 	      return cstring_copy (filename);
@@ -1110,10 +1147,11 @@ cstring osd_outputPath (cstring filename)
 
           *rel_buf_p++ = '.';
           *rel_buf_p++ = '.';
-          *rel_buf_p++ = '/';
+          *rel_buf_p++ = CONNECTCHAR;
         }
       
       /* Then tack on the unmatched part of the desired file's name.  */
+
       do
         {
           if (rel_buffer + filename_len <= rel_buf_p)
@@ -1124,12 +1162,14 @@ cstring osd_outputPath (cstring filename)
         } /*@-usereleased@*/
       while ((*rel_buf_p++ = *path_p++) != '\0') ;
 
-      /*@=usereleased@*/ /*@i523! shouldn't need these */
+      
+      /*@=usereleased@*/ /* Splint limitation: shouldn't need these */
       --rel_buf_p;
 
-      if (*(rel_buf_p-1) == '/')
+      if (osd_isConnectChar (*(rel_buf_p-1)))
         *--rel_buf_p = '\0';
 
+      /* fprintf (stderr, "Returning buffer: %s [%p]\n", rel_buffer, rel_buffer); */
       return rel_buffer;
     }
   /*@noaccess cstring@*/
